/***************************************************************************
                          keymapList.h  -  List of key mappings.
                             -------------------
    begin                : Sun Apr 8 2001
    copyright            : (C) 2001 by Lars Krueger
    email                : lars_e_krueger@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _keymapList_h_
#define _keymapList_h_
#ifdef _cplusplus
extern "C" {
#endif

#include <stdlib.h>
#include <regex.h>

/** @name Key map list handling.
  The defintions in this module deal with the key map lists.
*/
/*@{*/

#define KMI_LEFT        0
#define KMI_RIGHT       1
#define KMI_UP          2
#define KMI_DOWN        3

#define KMI_DIR_MAX     4

#define KMI_PUSH        4
#define KMI_RELEASE     5

#define KMI_MAX         6

/* Keep the constants in these order. Certain functions rely on it. */
#define KMOD_NONE       0
#define KMOD_BTN1       1
#define KMOD_BTN2       2
#define KMOD_BTN3       3
#define KMOD_BTN4       4
#define KMOD_BTN5       5
#define KMOD_SHIFT      6
#define KMOD_LOCK       7
#define KMOD_CTRL       8
#define KMOD_ALT        9
#define KMOD_MAX        10
  
typedef unsigned char tKeySym;

/** A single key map for one direction. */
typedef struct
{
  unsigned length;
  tKeySym * keys;
} tDirData;  

/** All the key maps for one application.*/
typedef struct keymapList_node
{
  struct keymapList_node * pNext;
  regex_t windowPattern;
  tDirData dirData[KMI_MAX][KMOD_MAX];
} keymapList_node;

/** Allocate a new set of key map for one applications. */
extern struct keymapList_node * keymapListAlloc();

/** Append a key map set to a list. */
extern void keymapListAppend( struct keymapList_node * list,
    struct keymapList_node * tail);

/*@}*/

#ifdef _cplusplus
}
#endif
#endif

