/***************************************************************************
                          lockFile.h  -  Lock file management
                             -------------------
    begin                : Sat Apr 7 2001
    copyright            : (C) 2001 by Lars Krueger
    email                : lars_e_krueger@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _lockFile_h_
#define _lockFile_h_

#include <unistd.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @name Lock file management
  These functions deal with the management of lock files that are used by daemons.
  The presence of a lock file denotes an already running daemon. A lock file
  contains the process id of the running daemon.

*/
/**@{*/

/** Setup the lock file processing. Call once before you call one of the other
 * functions in this header. */
void setupLockFile( const char * applicationName);

/** Kill the application that wrote the lock file. Remove the lock file if it
 * exists. Call only after validateLockFile returned true. */
void killLockFile();

/** Validate that there is no lock file. Return 1 if there is one. */
int validateLockFile();

/** Return the PID that was in the validated lock file. */
pid_t lockFilePid();

/** Return the name of the lock file. */
const char * lockFileName();

/** Create the lock file. */
void createLockFile();

/**@}*/

#ifdef __cplusplus
}
#endif
#endif

