/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : Sat Apr  7 19:39:27 BST 2001
    copyright            : (C) 2001 by Lars Krueger
    email                : lars_e_krueger@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/extensions/XTest.h>

#include "keymapList.h"
#include "lockFile.h"
#include "fopenFromList.h"

#ifdef USE_SMALL_CURSOR
#include "small_up_img.xbm"
#include "small_up_mask.xbm"

#include "small_down_img.xbm"
#include "small_down_mask.xbm"

#include "small_left_img.xbm"
#include "small_left_mask.xbm"

#include "small_right_img.xbm"
#include "small_right_mask.xbm"

#include "small_left_right_img.xbm"
#include "small_left_right_mask.xbm"

#include "small_up_down_img.xbm"
#include "small_up_down_mask.xbm"

#include "small_grab_img.xbm"
#include "small_grab_mask.xbm"
#else
#include "up_img.xbm"
#include "up_mask.xbm"

#include "down_img.xbm"
#include "down_mask.xbm"

#include "left_img.xbm"
#include "left_mask.xbm"

#include "right_img.xbm"
#include "right_mask.xbm"

#include "left_right_img.xbm"
#include "left_right_mask.xbm"

#include "up_down_img.xbm"
#include "up_down_mask.xbm"

#include "grab_img.xbm"
#include "grab_mask.xbm"
#endif

static keymapList_node * keymapList=0;
static Display * display;

#define INTERKEY_DELAY 0
#define POSTPUSH_DELAY 0

Cursor dirCursor[KMI_DIR_MAX];
Cursor grabCursorAll;
Cursor grabCursorHor;
Cursor grabCursorVer;

#define DEFAULT_TICKS 3
static unsigned int ticks = DEFAULT_TICKS;
#define DEFAULT_BUTTON_TO_GRAB Button4
static unsigned int buttonToGrab = DEFAULT_BUTTON_TO_GRAB;

static unsigned useLockFile=1;
static unsigned doFocus=1;
static unsigned wantToForward=0;

static void printHelp()
{
  puts(
      "fxred " VERSION " -- Scroll button daemon for the TrackMan Marble FX\n"
      "  usage: fxred [options]\n"
      "        -l        : Don't use the lock file\n"
      "        -f        : Don't focus windows\n"
      "        -k        : Kill the running instance (needs lockfile)\n"
      "        -q        : Quit before grabbing the red key\n"
      "        -d        : Don't become a daemon\n"
      "        -b <nr>   : Button that corresponds to the 'scrollwheel' mode. [4]\n"
      "        -t <ticks>: Emit key if moving further than <ticks>. [3]\n"
      "        -w        : Forward scroll button click without scrolling.\n"
      "        -h        : Show this help\n"
      );
}

static void processCommandLine( int argc, char *argv[])
{
  int becomeDaemon=1;
  int killLock=0;
  int quitNow=0;
  int running=1;
  while( running)
  {
    int opt=getopt( argc, argv, "lfkqdht:b:w");
    switch( opt)
    {
      case -1:
        running=0;
        break;
      case 'l':
        useLockFile=0;
        break;
      case 'f':
        doFocus=0;
        break;
      case 'k':
        killLock=1;
        break;
      case 'd':
        becomeDaemon=0;
        break;
      case 'b':
        buttonToGrab=atoi(optarg);
	break;
      case 'h':
        printHelp();
        exit(0);
        break;
      case 'q':
        quitNow=1;
        break;
      case 't':
	ticks=atoi(optarg);
        break;
      case 'w':
        wantToForward=1;
        break;
    }
  }
  if( useLockFile)
    setupLockFile( "fxred");
  if( killLock && useLockFile)
  {
    if( validateLockFile())
      killLockFile();
  }
  if( quitNow)
    exit(0);
  if( becomeDaemon)
  {
    pid_t fr=fork();
    if( fr==-1)
      perror( "fxred");
    else
    { 
      if( fr)
      {
        /* Parent */
        exit(0);
      }
      else
      {
        if( useLockFile)
        {
          /* Child */
          if( validateLockFile())
          {
            printf(
                "fxred: Already running as pid %d or stale lock file.\n"
                "       run 'fxred -k' or remove '%s'.\n",
                lockFilePid(), lockFileName());
            exit(1);
          }
          createLockFile();
        }
      }
    }
  }
}

static regex_t kmcRegex;

static void prepareKeymapCompilation()
{
  regcomp( &kmcRegex, "^([^|,]+)([|,]?)", REG_EXTENDED|REG_NEWLINE);
}

static void appendToDirData( tDirData * pDd, tKeySym ks)
{
  if( !pDd->keys)
  {
    pDd->keys=(tKeySym*)malloc(sizeof(tKeySym));
  }
  else
  {
    pDd->keys=(tKeySym*)realloc(pDd->keys, (pDd->length+1)*sizeof(tKeySym));
  }
  pDd->keys[pDd->length]=ks;
  pDd->length++;
}

static void compileKeymap( tDirData * pDd, char * keys, unsigned lineNr)
{
  regmatch_t match[3];
  unsigned lastKeyHadPipe=0;
  unsigned countInd=0;
  pDd->length=0;
  pDd->keys=0;
  while( *keys)
  {
    if( !regexec( &kmcRegex, keys, 3, match, 0))
    {
      KeySym symRead;
      KeyCode codeRead;
      unsigned thisKeyHasPipe;

      thisKeyHasPipe=((match[2].rm_so!=-1) && (keys[match[2].rm_so]=='|'));
      keys[match[1].rm_eo]=0;

      {
        symRead=XStringToKeysym( keys);
        if( symRead==NoSymbol)
        {
          printf( "fxred: Unknown keysym '%s' in line %d.\n"
              "       Ignoring line.\n",keys,lineNr);
          if( pDd->keys)
            free(pDd->keys);
          pDd->length=0;
          return;
        }
        codeRead=XKeysymToKeycode( display, symRead);
        if( lastKeyHadPipe)
        {
          appendToDirData( pDd, codeRead);
          pDd->keys[countInd]++;
        }
        else
        {
          countInd=pDd->length;
          appendToDirData( pDd, 1);
          appendToDirData( pDd, codeRead);
        }
      }
      lastKeyHadPipe=thisKeyHasPipe;
      keys+=match[0].rm_eo;
    }
    else
    {
      printf("fxred: Keymap list syntax error in line %d after '%s'.\n"
             "       Ignoring line.\n",
          lineNr, keys);
      if( pDd->keys)
        free(pDd->keys);
      pDd->length=0;
      return;
    }
  }
}

static void finishKeymapCompilation()
{
  regfree( &kmcRegex);
}

static void grabButton( Cursor curs)
{
  if( !XGrabButton(
      display,
      buttonToGrab,
      0,
      DefaultRootWindow(display),
      False,
      ButtonReleaseMask|ButtonPressMask|PointerMotionMask,
      GrabModeAsync, GrabModeAsync,
      None,
      curs))
  {
    printf( "fxred: Could not grab red button.\n"
            "       Leaving now.\n");
    exit(1);
  }
}

static void ungrabButton()
{
  XUngrabButton( display, buttonToGrab, 0, DefaultRootWindow( display));
}

static void changeGrab( Cursor curs)
{
  XChangeActivePointerGrab( display, 
      ButtonReleaseMask|ButtonPressMask|PointerMotionMask, 
      curs, CurrentTime);
}

static void executeDirData( const tDirData * pDd)
{
  int ind,keyNr,keyCnt;
  XTestGrabControl( display,True);
  for( ind=0; ind<pDd->length; ind++)
  {
    keyCnt=pDd->keys[ind];
    for( keyNr=0; keyNr<keyCnt; keyNr++)
    {
      XTestFakeKeyEvent( display, pDd->keys[ind+1+keyNr],True,INTERKEY_DELAY);
      XSync(display, False);
    }
    usleep( POSTPUSH_DELAY);
    for( ; --keyNr>=0; )
    {
      XTestFakeKeyEvent( display, pDd->keys[ind+1+keyNr],False,INTERKEY_DELAY);
      XSync(display, False);
    }
    ind+=keyCnt;
  }
  XTestGrabControl( display, False);
  XSync(display, False);
}

/**
 * skip whitespaces at the beginning of the string
 * and return a pointer at the 1st non-whitespace character (or '\0')
 */
static char *strtrim(char *str)
{
  while(*str==' ' || *str=='\t')
    str++;
  return str;
}

static void readConfiguration()
{
  FILE * fi;
  char homeRc[FILENAME_MAX];
  struct keymapList_node ** lastNode=&keymapList;
  struct keymapList_node * currNode=0;
  unsigned lineNr=0;

  strcpy( homeRc, getenv("HOME"));
  strcat( homeRc, "/.fxredrc");
  
  fi=fopenFromList("r", 
      "fxredrc", 
      "../fxredrc", 
      homeRc, 
      PREFIX "/etc/fxredrc", 
      NULL);
  if( !fi)
  {
    printf( "fxred: Can't open configuration file.\n");
    exit(1);
  }
  prepareKeymapCompilation();
  while( !feof(fi))
  {
    char windowPatternString[256];
    char keys[256];
    char dir[2];
    char modifier[2];
    int res;
    long lineStartPos=ftell(fi);
    lineNr++;
    /* Skip empty lines. */
    res=fscanf( fi, "%*1[\n]");
    if( lineStartPos!=ftell(fi))
      continue;
    /* Skip comment lines. */
    res=fscanf( fi, "#%*[^\n]%*c");
    if( lineStartPos!=ftell(fi))
      continue;
    if( feof(fi))
      break;
    if( (res=fscanf( fi, "\"%255[^\"]\"\n", windowPatternString))==1)
    {/* Check for application section */
      /* New window name pattern. Compile regex. */
      regex_t re;
      if( regcomp( &re, windowPatternString, REG_EXTENDED|REG_NOSUB))
      {
        printf("fxred: Illegal regex in configuration file, line %d.\n"
               "       Ignoring rest of file.\n", lineNr);
        break;
      }
      currNode=keymapListAlloc();
      currNode->windowPattern=re;
      *lastNode=currNode;
      lastNode=&currNode->pNext;
    }
    else if( (res=fscanf( fi, "%1[lrud><]%1[1-5sScClLaA ]%255[^\n]%*1c", dir, modifier, keys))==3)
    {/* Check for event */
      int ind=KMI_LEFT;
      int imod;
      if( currNode)
      {
        switch( dir[0])
        {
          case 'l': ind=KMI_LEFT;         break;
          case 'r': ind=KMI_RIGHT;        break;
          case 'u': ind=KMI_UP;           break;
          case 'd': ind=KMI_DOWN;         break;
          case '<': ind=KMI_RELEASE;      break;
          case '>': ind=KMI_PUSH;         break;
        }
        switch( modifier[0])
        {
          case '1': imod=KMOD_BTN1;       break;
          case '2': imod=KMOD_BTN2;       break;
          case '3': imod=KMOD_BTN3;       break;
          case '4': imod=KMOD_BTN4;       break;
          case '5': imod=KMOD_BTN5;       break;
          case 's':
          case 'S': imod=KMOD_SHIFT;      break;
          case 'c':
          case 'C': imod=KMOD_CTRL;       break;
          case 'l':
          case 'L': imod=KMOD_LOCK;       break;
          case 'a':
          case 'A': imod=KMOD_ALT;        break;
          default:  imod=KMOD_NONE;       break;
        }
        compileKeymap( &currNode->dirData[ind][imod], strtrim(keys), lineNr);
      }
      else
      {
        printf("fxred: Found illegal direction in line %d.\n"
               "       Ignored.\n",
            lineNr);
      }
    }
    else
    {
      printf("fxred: Error in configuration file, line %d.\n"
             "       Ignoring rest of file.\n", lineNr);
      break;
    }
  }
  finishKeymapCompilation();
  fclose(fi);
}

static void openDisplay()
{
  display=XOpenDisplay(0);
  if(!display)
  {
    printf("fxred: Could not open display. Check shell DISPLAY variable\n"
           "       and export or setenv it!\n");
    exit(1);
  }
}

static Cursor makeCursor( unsigned char * img, unsigned char * mask, unsigned int w, unsigned int h, const char *white, const char *black)
{
  XColor fg,bg;
  Pixmap imgPm, maskPm;
  Cursor res;

  if(!white)
    white="White";
  if(!black)
    black="Black";

  XParseColor( display, DefaultColormap(display,DefaultScreen(display)),
    white, &bg);
  XParseColor( display, DefaultColormap(display,DefaultScreen(display)),
    black, &fg);

  imgPm=XCreateBitmapFromData( display, DefaultRootWindow(display), img,
      w,h);
  maskPm=XCreateBitmapFromData( display, DefaultRootWindow(display), mask,
      w,h);

  res=XCreatePixmapCursor( display, imgPm, maskPm, &fg, &bg, w/2, h/2);
  XFreePixmap( display, imgPm);
  XFreePixmap( display, maskPm);

  return res;
}

static void prepareX()
{
  int i;

  XAllowEvents(display,AsyncBoth,CurrentTime);
  XAllowEvents(display,SyncBoth,CurrentTime);
  for( i=0; i<KMI_DIR_MAX; i++)
    dirCursor[i]=None;
#define MAKE_CURSOR(name, fg, bg) makeCursor( name ## _img_bits, name ## _mask_bits, name ## _img_width, name ## _img_height, fg, bg ) 
#define MAKE_STD_CURSOR(name) MAKE_CURSOR(name, NULL/*white*/, NULL/*black*/)
#define MAKE_GRAB_CURSOR(name) MAKE_CURSOR(name, NULL/*white*/, "#777777")

  dirCursor[KMI_UP]=    MAKE_STD_CURSOR(up) ;
  dirCursor[KMI_LEFT]=  MAKE_STD_CURSOR(left);
  dirCursor[KMI_DOWN]=  MAKE_STD_CURSOR(down);
  dirCursor[KMI_RIGHT]= MAKE_STD_CURSOR(right); 

  grabCursorAll=MAKE_GRAB_CURSOR(grab);
  grabCursorHor=MAKE_GRAB_CURSOR(left_right);
  grabCursorVer=MAKE_GRAB_CURSOR(up_down);
  
  grabButton( grabCursorAll);
}

/* Return the keymap of the application. */
static struct keymapList_node * findNodeFromString( int dummy, ...)
{
  va_list va;
  struct keymapList_node * res=0;
  struct keymapList_node * node=keymapList;

  while( node)
  {
    va_start( va, dummy);
    while( 1)
    {
      const char * name=va_arg( va, const char * );
      if( !name)
        break;
      if( !regexec( &node->windowPattern, name, 0, 0, 0))
      {
        /* Name found. */
        res=node;
        va_end(va);
        return res;
      }
    }
    va_end(va);
    node=node->pNext;
  }
  return 0;
}

static struct keymapList_node * findNodeToRun( const char * windowName, const char * rn,
    const char * rc, Status sfn, Status sgch)
{
  if( sfn && sgch)
    return findNodeFromString( 0, windowName, rn, rc, 0);
  else if( sfn)
    return findNodeFromString( 0, windowName, 0);
  else if( sgch)
    return findNodeFromString( 0, rn, rc, 0);
  return 0;
}

static void freeWindowNames( Status sfn, Status sgch, char * windowName, 
    XClassHint * xclasshint)
{
  if( sfn)
    XFree( windowName);
  if( sgch)
  {
    XFree( xclasshint->res_class);
    XFree( xclasshint->res_name);
  }
}

/* This is a tricky one. Some window manager do not report the real window but
 * one of its parents. As we assume not to be called for the root window,
 * traversing the child tree shouldn't take long. 
 *
 * We return 1 if we found a named window .*/
static unsigned getWindowNames( Window win, Status * sfn, Status * sgch, char ** windowName, XClassHint * xclasshint)
{
  Window rw, pw, *cw;
  unsigned nchildren, i, found;
  *sfn=XFetchName(display, win,windowName);
  *sgch=XGetClassHint(display,win,xclasshint);
  if( *sfn || *sgch)
  {
    if( 
        strcmp( *windowName, "kwin") &&
        strcmp( xclasshint->res_name, "unnamed"))
    {
      return 1; 
    }
    freeWindowNames( *sfn, *sgch, *windowName, xclasshint);
  }
  /* Get the children and and traverse them depth first. */
  if( !XQueryTree(display, win, &rw,&pw,&cw, &nchildren))
    return 0;
  found=0;
  for( i=0; (i<nchildren) && !found; i++)
  {
    if( getWindowNames( cw[i], sfn, sgch, windowName, xclasshint))
    {
      found=1;
      break;
    }
  }
  XFree( cw);
  return found;
}


static unsigned motionDirection(int dx, int dy)
{
#define dx_is_positive dx>0 && dx>abs(dy) 
#define dx_is_negative dx<0 && -dx>abs(dy)
#define dy_is_positive dy>0 && dy>abs(dx)
#define dy_is_negative dy<0 && -dy>abs(dx)

  if( dx_is_positive)
      return KMI_RIGHT;

  if( dx_is_negative)
      return KMI_LEFT;

  if( dy_is_negative)
      return KMI_UP;

  if( dy_is_positive)
      return KMI_DOWN;

  return -1;
}

static int hasDirData(const tDirData *dirData)
{
  int imod;
  for(imod=0; imod<KMOD_MAX; imod++)
    if(dirData[imod].length!=0)
      return 1;
  return 0;
}

static unsigned modifierMask[]={
Button1Mask, /* KMOD_BTN1     */
Button2Mask, /* KMOD_BTN2     */
Button3Mask, /* KMOD_BTN3     */
Button4Mask, /* KMOD_BTN4     */
Button5Mask, /* KMOD_BTN5     */
ShiftMask,   /* KMOD_SHIFT    */
LockMask,    /* KMOD_LOCK     */
ControlMask, /* KMOD_CTRL     */
Mod1Mask,    /* KMOD_ALT      */
};
/**
 * Take a 'button state' as found in an xmotion xevent, for example
 * and return one of the KMOD_ constants.
 * 
 * Only the 1st pressed button, is returned, excluding buttonToGrab.
 */
static unsigned int stateToModifier(unsigned int state)
{
  unsigned int i;
  for(i=1; i<KMOD_MAX; i++)
  {
    if(i==buttonToGrab)
      continue;
    if( (state& modifierMask[i-1]) != 0 )
      return i;
  }
  return 0;
}

/**
 * Send button press and release event 
 */
static void sendButtonPressEvents(int button)
{
  XTestGrabControl(display, True);
  XTestFakeButtonEvent(display, button, True/*press*/, 0);
  XTestFakeButtonEvent(display, button, False/*release*/, 0);
  XTestGrabControl(display, False);
  XSync(display, False);
}

static void processEvents()
{
  XEvent e;
  Window dw, drw, cw;
  Status sfn, sgch;
  int di;
  unsigned du;
  int dx, dy;
  XClassHint xclasshint;
  char * windowName;
  struct keymapList_node * nodeToRun=NULL;
  int downX, downY;
  int lastdir=-1;
  int dircount=0;
  int executed=0/*false*/; // set to 0 at buttonDown, 1 when something is executed, checked at buttonRelease

  Window oldFocus;
  int oldRevertTo;

  //XSetErrorHandler((XErrorHandler)nullXError);
  drw=DefaultRootWindow(display);
  while(1)
  {
    XNextEvent(display,&e);
    switch(e.type)
    {
      case ButtonPress:
        XQueryPointer( display, drw, &dw, &cw,
            &downX, &downY,
            &di, &di, &du);
        if( e.xbutton.button==buttonToGrab && e.xbutton.subwindow)
        {
          if( doFocus)
          {
            XGetInputFocus( display, &oldFocus, &oldRevertTo); 
            XSetInputFocus( display, e.xbutton.subwindow, RevertToNone, CurrentTime); 
            XSync( display, False); 
          }

          getWindowNames( e.xbutton.subwindow, &sfn, &sgch, &windowName, &xclasshint);
          nodeToRun=findNodeToRun( windowName, xclasshint.res_name,
              xclasshint.res_class, sfn, sgch);
          if( nodeToRun)
          {
            int flags=0;
            executeDirData( &nodeToRun->dirData[KMI_PUSH][0]);
            if( hasDirData(nodeToRun->dirData[KMI_LEFT]) &&
                hasDirData(nodeToRun->dirData[KMI_RIGHT]))
              flags=1;
            if( hasDirData(nodeToRun->dirData[KMI_UP]) &&
                hasDirData(nodeToRun->dirData[KMI_DOWN]))
              flags|=2;
            switch( flags)
            {
              case 0: /* Empty application pattern. */
                break;
              case 1: /* horizontal only */
                changeGrab( grabCursorHor);
                break;
              case 2: /* Vertical */
                changeGrab( grabCursorVer);
                break;
              case 3: /* both */
                changeGrab( grabCursorAll);
                break;
            }
          }
          freeWindowNames( sfn, sgch, windowName, &xclasshint);

          lastdir=-1;
	  executed=0/*false*/;
        }
        break;
      case MotionNotify:
        XWarpPointer( display, None, drw, 0,0,0,0, downX, downY);
        if( nodeToRun)
        {
          dx=e.xmotion.x-downX;
          dy=e.xmotion.y-downY;
          if( dx || dy)
          {
            if( e.xmotion.subwindow)
            {
              const unsigned int dir = motionDirection(dx, dy);
              if(dir==lastdir)
                dircount++;
              else
                dircount=0;
              lastdir=dir;

              getWindowNames( e.xmotion.subwindow, &sfn, &sgch, &windowName, &xclasshint);
              if( dir!=-1)
              {
                const unsigned int imod = stateToModifier(e.xmotion.state);
                const tDirData *dirData = &nodeToRun->dirData[dir][imod];

                if(dirData->length==0)
                  dirData=&nodeToRun->dirData[dir][0]; /* use default, w/o modifiers */

		
		if(dirData->length>0)
		  {
		    if(dircount>=ticks)
		      {
			executeDirData( dirData );
			dircount=0;
		      }
		    changeGrab( dirCursor[dir]);
		  }
              }
              freeWindowNames( sfn, sgch, windowName, &xclasshint);
            }
          }
        }
        break;
      case ButtonRelease:
        if( nodeToRun && e.xbutton.button==buttonToGrab && e.xbutton.subwindow)
        {
          getWindowNames( e.xbutton.subwindow, &sfn, &sgch, &windowName, &xclasshint);
          executeDirData( &nodeToRun->dirData[KMI_RELEASE][0]);
          freeWindowNames( sfn, sgch, windowName, &xclasshint);

          if( doFocus)
          {
            XSetInputFocus( display, oldFocus, oldRevertTo, CurrentTime); 
            XSync( display, False);
          }

          if( !executed && wantToForward)
          {
            ungrabButton();
            sendButtonPressEvents(buttonToGrab);
            grabButton(grabCursorAll);
          }
        }
        break;

    }
  } // infinite loop
}


int main(int argc, char *argv[])
{
  processCommandLine( argc, argv);
  openDisplay();
  readConfiguration();
  prepareX();
  atexit( ungrabButton);
  processEvents();

  return EXIT_SUCCESS;
}
